/* Bilateralフィルタ */
/* グレースケールの標準偏差を計算する関数 */
double getStandardDeviation(Image3D *R,Image3D *G,Image3D *B){
  double ave=0.0;
  double sd=0.0;
  int i,j,k;
  for(k=0;k<R->st;k++)
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      double gray = (R->img[k][i][j]+G->img[k][i][j]+B->img[k][i][j])/3.0;
      ave += gray;
      sd += gray*gray;
    }
  ave /= ((double)(R->sx*R->sy*R->st));
  sd /= ((double)(R->sx*R->sy*R->st));
  sd -= ave*ave;
  sd = sqrt(sd);
  return sd;

}
void setBilateral(Image3D *outR,Image3D *outG,Image3D *outB,Image3D *tmpsum,Image3D *R,Image3D *G,Image3D *B,int ti,int tj,int tk,int radii,int radii_t,double *wXY,double *wT,double *wI,int size,double sigI){
  
  int i,j,k;
  int ib=ti+radii_t;if(ib>=R->st)ib=R->st-1;
  
  int jbm=tj-radii;if(jbm<0)jbm=0;
  int jb=tj+radii;if(jb>=R->sy)jb=R->sy-1;
  
  int kbm=tk-radii;if(kbm<0)kbm=0;
  int kb=tk+radii;if(kb>=R->sx)kb=R->sx-1;
  double invalR = R->img[ti][tj][tk];
  double invalG = G->img[ti][tj][tk];
  double invalB = B->img[ti][tj][tk];
  
  double xw,yw,zw,weight;
  
  i = ti;
  j = tj;
  for(k=tk+1;k<=kb;k++){
    xw = wXY[abs((k-tk))];
    double difR = R->img[i][j][k]-invalR;
    double difG = G->img[i][j][k]-invalG;
    double difB = B->img[i][j][k]-invalB;
    //double grad = ((difR*difR+difG*difG+difB*difB)/3.0);
    //weight =  xw*exp(-(grad*sigI));
    double grad = (fabs(difR)+fabs(difG)+fabs(difB))/3.0;
    int grad_i = ((int)(grad));if(fabs(grad_i-grad)>=0.5)grad_i++;
    
    if(grad_i>=size)grad_i=size-1;
    weight =  xw*wI[grad_i];
    tmpsum->img[ti][tj][tk] += weight;
    tmpsum->img[i][j][k] += weight;
    outR->img[ti][tj][tk] += weight*R->img[i][j][k];
    outR->img[i][j][k] += weight*R->img[ti][tj][tk];
    outG->img[ti][tj][tk] += weight*G->img[i][j][k];
    outG->img[i][j][k] += weight*G->img[ti][tj][tk];
    outB->img[ti][tj][tk] += weight*B->img[i][j][k];
    outB->img[i][j][k] += weight*B->img[ti][tj][tk];
  }
  
  i= ti;
  for(j=tj+1;j<=jb;j++){
    yw = wXY[abs((j-tj))];
    for(k=kbm;k<=kb;k++){
      xw = wXY[abs((k-tk))];
      double difR = R->img[i][j][k]-invalR;
      double difG = G->img[i][j][k]-invalG;
      double difB = B->img[i][j][k]-invalB;
      
      //double grad = ((difR*difR+difG*difG+difB*difB)/3.0);
      double grad = (fabs(difR)+fabs(difG)+fabs(difB))/3.0;
      int grad_i = ((int)(grad));if(fabs(grad_i-grad)>=0.5)grad_i++;if(grad_i>=size)grad_i=size-1;
      weight =   yw*xw*wI[grad_i];
      //weight =   yw*xw*exp(-(grad*sigI));
      tmpsum->img[ti][tj][tk] += weight;
      tmpsum->img[i][j][k] += weight;
      outR->img[ti][tj][tk] += weight*R->img[i][j][k];
      outR->img[i][j][k] += weight*R->img[ti][tj][tk];
      outG->img[ti][tj][tk] += weight*G->img[i][j][k];
      outG->img[i][j][k] += weight*G->img[ti][tj][tk];
      outB->img[ti][tj][tk] += weight*B->img[i][j][k];
      outB->img[i][j][k] += weight*B->img[ti][tj][tk];
    }
  }
  
  for(i=ti+1;i<=ib;i++){
    zw = wT[abs((i-ti))];
    for(j=jbm;j<=jb;j++){
      yw = wXY[abs((j-tj))]*zw;
      for(k=kbm;k<=kb;k++){
	double difR = R->img[i][j][k]-invalR;
	double difG = G->img[i][j][k]-invalG;
	double difB = B->img[i][j][k]-invalB;
	//double grad = ((difR*difR+difG*difG+difB*difB)/3.0);
	double grad = (fabs(difR)+fabs(difG)+fabs(difB))/3.0;
	int grad_i = ((int)(grad));if(fabs(grad_i-grad)>=0.5)grad_i++;if(grad_i>=size)grad_i=size-1;
	xw = wXY[abs((k-tk))];
	weight =  yw*xw*wI[grad_i];
	//weight =  yw*xw*exp(-(grad*sigI));
	//printf("%lf %lf %lf\n",yw,xw,weight);
	tmpsum->img[ti][tj][tk] += weight;
	tmpsum->img[i][j][k] += weight;
	outR->img[ti][tj][tk] += weight*R->img[i][j][k];
	outR->img[i][j][k] += weight*R->img[ti][tj][tk];
	outG->img[ti][tj][tk] += weight*G->img[i][j][k];
	outG->img[i][j][k] += weight*G->img[ti][tj][tk];
	outB->img[ti][tj][tk] += weight*B->img[i][j][k];
	outB->img[i][j][k] += weight*B->img[ti][tj][tk];
      }
    }
  }
  
}

void Copy(Image3D *R,Image3D *G,Image3D *B,Image3D *outR,Image3D *outG,Image3D *outB){
  int i,j,k;
  for(k=0;k<R->st;k++)
    for(i=0;i<R->sy;i++)
      for(j=0;j<R->sx;j++){
	outR->img[k][i][j] = R->img[k][i][j];
	outG->img[k][i][j] = G->img[k][i][j];
	outB->img[k][i][j] = B->img[k][i][j];
      }

}


 void IterativeBilateralFilter(int N,Image3D *R,Image3D *G,Image3D *B,Image3D *outR,Image3D *outG,Image3D *outB,int radii,int radii_t,double sigma_s,double sigma_t,double sigma_I){
   Image3D *tmpR = new Image3D(R->sx,R->sy,R->st);
   Image3D *tmpG = new Image3D(G->sx,G->sy,G->st);
   Image3D *tmpB = new Image3D(B->sx,B->sy,B->st);
   Image3D *tmpsum =  new Image3D(R->sx,R->sy,R->st);
   int i,j,k,n;
   
   int size = 128+1;
   
   
   double *gauss_intensity = new double[size];
   
   
   double sigS = 1.0/(2.0*sigma_s*sigma_s);
   double sigT = 1.0/(2.0*sigma_t*sigma_t);
  
   double *wXY = new double[(radii+1)];
   double *wT = new double[(radii_t+1)];
   for(i=0;i<(radii+1);i++){
     wXY[i] = exp(-((double)(i*i))*sigS);
   }
   for(i=0;i<(radii_t+1);i++){
     wT[i] = exp(-((double)(i*i))*sigT);
  }

   
   
   if(N==0){
     Copy(R,G,B,outR,outG,outB);
     
   }else{
     Copy(R,G,B,tmpR,tmpG,tmpB);
     
     for(n=0;n<N;n++){ 
       printf("BilateralFilter ite = %d\n",n);
       
     double standard_deviation = getStandardDeviation(tmpR,tmpG,tmpB);
     double sigma_i =sigma_I*standard_deviation; 
     /* 効率化のためGauss関数の中でsigIをかけるだけにするための変数 */
     double sigI = 1.0/(2.0*sigma_i*sigma_i);
     
     for(i=0;i<size;i++)gauss_intensity[i] = exp(-((double)(i*i))*sigI);
     
     for(i=0;i<tmpsum->st;i++)
       for(j=0;j<tmpsum->sy;j++)
	 for(k=0;k<tmpsum->sx;k++)
	   tmpsum->img[i][j][k] = 0.0;
     for(i=0;i<tmpsum->st;i++)
       for(j=0;j<tmpsum->sy;j++)
	 for(k=0;k<tmpsum->sx;k++){
	   outR->img[i][j][k]=tmpR->img[i][j][k];
	   outG->img[i][j][k]=tmpG->img[i][j][k];
	   outB->img[i][j][k]=tmpB->img[i][j][k];
	   tmpsum->img[i][j][k]=1.0;
	 }
     for(i=0;i<tmpsum->st;i++){
       printf("Bilateral Filtering time = %d\n",i);
       	
	for(j=0;j<tmpsum->sy;j++)
	  for(k=0;k<tmpsum->sx;k++)
	    setBilateral(outR,outG,outB,tmpsum,tmpR,tmpG,tmpB,i,j,k,radii,radii_t,wXY,wT,gauss_intensity,size,sigI); 
	
     }
     for(i=0;i<tmpsum->st;i++)
       for(j=0;j<tmpsum->sy;j++)
	 for(k=0;k<tmpsum->sx;k++){
	   outR->img[i][j][k] /= tmpsum->img[i][j][k];
	   outG->img[i][j][k] /= tmpsum->img[i][j][k];
	   outB->img[i][j][k] /= tmpsum->img[i][j][k];
	 }
     
     if(n>=N-1)break;
     Copy(outR,outG,outB,tmpR,tmpG,tmpB);
     
   }
   }
   
   delete [] wXY;
   delete [] wT;
   delete [] gauss_intensity;
   delete tmpR;
   delete tmpG;
   delete tmpB;
   delete tmpsum;
   
}
