#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include <vector> 
#include <algorithm>
#include"SimpleImage.h"
#include"BMPIO.h"
#include"SimpleImage3D.h"
#include"VideoIO.h"
#include"Bilateral.h"

#define PI 3.141592653589793238462643383279502884197


/* DoG画像を生成 */
Image3D* makeDoGImage(int radii,int radii_t,double sig,double sig_t,double K){
  int gsx = 2*radii+1;
  int gsy = gsx;
  int gst= 2*radii_t+1;
  double sig2 = 2.0*sig*sig;
  double sigUV1 = 1.0/sig2;
  double sigUV2 = 1.0/(sig2*K*K);
  double coeff1 = 1.0/(sig2*PI);
  double coeff2 = 1.0/(sig2*PI*K*K);
  double sigUV1_t = 1.0/(2.0*sig_t*sig_t);
  double sigUV2_t = 1.0/(2.0*sig_t*sig_t*K*K);
  double coeff1_t = 1.0/(sqrt(2.0*PI)*sig_t);
  double coeff2_t = 1.0/(sqrt(2.0*PI)*sig_t*K);
  
  
  Image3D *tmp_img = new Image3D(gsx,gsy,gst);
  int i,j,k;
 
  for(k=-radii_t;k<=radii_t;k++){
    int w = k+radii_t; int k2 = k*k;
    for(i=-radii;i<=radii;i++){
      int v = i+radii; int i2 = i*i;
      for(j=-radii;j<=radii;j++){
	int u = j+radii;
	double dist = ((double)(i2+j*j));
	double dist_t = ((double)(w*w));
	
	 tmp_img->img[w][v][u] = coeff1_t*exp(-(dist_t*sigUV1_t))*(coeff1*exp(-(dist*sigUV1))-coeff2*exp(-(dist*sigUV2)));
	
      }
    }
  }
  
  return tmp_img;
}

/* 畳み込み */
 void Conv(Image3D *in,int radii,int radii_t,Image3D *kernel,Image3D *out){
  int i,j,k;
  for(k=0;k<in->st;k++)
    for(i=0;i<in->sy;i++)
      for(j=0;j<in->sx;j++){
	int k2,i2,j2;
	out->img[k][i][j] = 0.0;
	for(k2=-radii_t;k2<=radii_t;k2++){
	  int w = k2+radii_t;
	  int w2 = k+k2;if(w2<0)w2=0;if(w2>in->st-1)w2=in->st-1;
	  for(i2=-radii;i2<=radii;i2++){
	    int v = i2+radii;
	    int v2 = i+i2;if(v2<0)v2=0;if(v2>in->sy-1)v2=in->sy-1;
	    for(j2=-radii;j2<=radii;j2++){
	      int u = j2+radii;
	      int u2 = j+j2;if(u2<0)u2=0;if(u2>in->sx-1)u2=in->sx-1;
	      out->img[k][i][j] += in->img[w2][v2][u2]*kernel->img[w][v][u];
	      
	    }
	  }
	  
	}
      }
  }

/* 正規化用の関数 */
void Normalize(Image3D *R,Image3D *G,Image3D *B){
  double max,min;
  max= min = R->img[0][0][0];
  int i,j,k;
  for(k=0;k<R->st;k++)
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      if(max<R->img[k][i][j])max=R->img[k][i][j];
      if(max<G->img[k][i][j])max=G->img[k][i][j];
      if(max<B->img[k][i][j])max=B->img[k][i][j];
      if(min>R->img[k][i][j])min=R->img[k][i][j];
      if(min>G->img[k][i][j])min=G->img[k][i][j];
      if(min>B->img[k][i][j])min=B->img[k][i][j];
    }
  double range = fabs(max-min);if(range<=0.0)range=1.0;
   for(k=0;k<R->st;k++)
  for(i=0;i<R->sy;i++)
     for(j=0;j<R->sx;j++){
       double val = 255.0*(R->img[k][i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      R->img[k][i][j] = val;
      val = 255.0*(G->img[k][i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      G->img[k][i][j] = val;
      val = 255.0*(B->img[k][i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      B->img[k][i][j] = val;
     }
     
}


/* 量子化 */
void Quantization(Image3D *in,Image3D *out,int binN){
  int i,j,k;
  double max=in->img[0][0][0];
  for(k=0;k<in->st;k++)
    for(i=0;i<in->sy;i++)
      for(j=0;j<in->sx;j++){
	if(max<in->img[k][i][j])max=in->img[k][i][j];
      }
  if(max<=0.0)max=1.0;
  for(k=0;k<in->st;k++)
    for(i=0;i<in->sy;i++)
      for(j=0;j<in->sx;j++){
	double val = ((double)(binN))*in->img[k][i][j]/max;
	int vali = ((int)(val));
	if(fabs(((double)(vali))-val)>=0.5)vali++;
	out->img[k][i][j] = ((double)(vali))*max/((double)(binN));
      }
  }
/* 量子化 */
void Quantization(Image3D *in,Image3D *out,int radii_t,int binN){
  int i,j,k,t;
  
  
  for(t=0;t<in->st;t++){
    
    double max=in->img[t][0][0];
    
    for(k=t-radii_t;k<=t+radii_t;k++){
      if(k>=0&&k<in->st)
	for(i=0;i<in->sy;i++)
	  for(j=0;j<in->sx;j++){
	    if(max<in->img[k][i][j])max=in->img[k][i][j];
	  }
    }
    if(max<=0.0)max=1.0;
    for(i=0;i<in->sy;i++)
      for(j=0;j<in->sx;j++){
	double val = ((double)(binN))*in->img[t][i][j]/max;
	int vali = ((int)(val));
	if(fabs(((double)(vali))-val)>=0.5)vali++;
	out->img[t][i][j] = ((double)(vali))*max/((double)(binN));
      }
  }
}
/* maxで割る正規化 */
void Normalize(Image3D *in){
  int i,j,k;
  double min,max;
  min=max=in->img[0][0][0];
  for(k=0;k<in->st;k++)
    for(i=0;i<in->sy;i++)
      for(j=0;j<in->sx;j++){
	if(min>in->img[k][i][j])min=in->img[k][i][j];
	if(max<in->img[k][i][j])max=in->img[k][i][j];
      }
  max=fabs(max);
  if(max<=0.0)max=1.0;
  for(k=0;k<in->st;k++)
    for(i=0;i<in->sy;i++)
      for(j=0;j<in->sx;j++)in->img[k][i][j]/=max;
  
}


/* HSV量子化: Hue色相だけbinNの数に量子化、alphaは明度Vの強調パラメータ */
 void QuantizationHSV(Image3D *R,Image3D *G,Image3D *B,int binN,int radii_t,double alpha){
   int i,j,k;
   
  Normalize(R,G,B);
  
  int localst = 2*radii_t+1;
  //RGB->HSV
  for(k=0;k<R->st;k++){
    Image3D *H = new Image3D(R->sx,R->sy,localst);
    Image3D *S = new Image3D(R->sx,R->sy,localst);
    Image3D *V = new Image3D(R->sx,R->sy,localst);
    for(int l=-radii_t;l<=radii_t;l++){
      int l2 = l+radii_t;
      int w = l+k;if(w<0)w=0;if(w>=R->st)w=R->st-1;
      
      for(i=0;i<R->sy;i++)
	for(j=0;j<R->sx;j++){
	  double max,min;
	  int checkmax=0;
	  min = max = R->img[w][i][j];
	  if(max<G->img[w][i][j]){
	    max=G->img[w][i][j];
	    checkmax=1;
	  }
	  if(max<B->img[w][i][j]){
	    max=B->img[w][i][j];
	    checkmax=2;
	  }
	  if(min>G->img[w][i][j])min=G->img[w][i][j];
	  if(min>B->img[w][i][j])min=B->img[w][i][j];
	  
	  S->img[l2][i][j] = (max-min);
	  if(max>0.0)S->img[l2][i][j]/=max;
	  V->img[l2][i][j] = max;
	  
	  double div=max-min;
	  if(div<=0.0)div=1.0;
	  if(checkmax==0){
	    H->img[l2][i][j] = 60.0*(G->img[w][i][j]-B->img[w][i][j])/div;
	  }else if(checkmax==1){
	    H->img[l2][i][j] = 60.0*(B->img[w][i][j]-R->img[w][i][j])/div+120.0;
	  }else{
	    H->img[l2][i][j] = 60.0*(R->img[w][i][j]-G->img[w][i][j])/div+240.0;
	  }
	  if(H->img[l2][i][j]<0.0)H->img[l2][i][j]+=360.0;
	  
	}
    }
    Normalize(S);
    Normalize(V);
    for(int l=-radii_t;l<=radii_t;l++){
      int l2 = l+radii_t;
      for(i=0;i<R->sy;i++)
	for(j=0;j<R->sx;j++){
	  V->img[l2][i][j] =  V->img[l2][i][j]*20.0-10.0;
	  V->img[l2][i][j] = 255.0/(1.0+exp(-alpha*V->img[l2][i][j]));
	  
	}
    }
    Quantization(H,H,binN);
    
    
      //HSV->RGB
    for(i=0;i<R->sy;i++)
      for(j=0;j<R->sx;j++){
	  
	if(S->img[radii_t][i][j]==0.0){
	  R->img[k][i][j]=G->img[k][i][j]=B->img[k][i][j]=V->img[radii_t][i][j];
	}else{
	    double Hi = H->img[radii_t][i][j]/60.0;
	    int Hii = ((int)(Hi));
	    Hii = Hii % 6;
	    
	    double f = H->img[radii_t][i][j]/60.0-((double)(Hii));
	    double p = V->img[radii_t][i][j]*(1.0-S->img[radii_t][i][j]);
	    double q = V->img[radii_t][i][j]*(1.0-f*S->img[radii_t][i][j]);
	    double t = V->img[radii_t][i][j]*(1.0-(1.0-f)*S->img[radii_t][i][j]);
	    if(Hii==0){
	      R->img[k][i][j]=V->img[radii_t][i][j];G->img[k][i][j]=t;B->img[k][i][j]=p;
	    }else if(Hii==1){
	      R->img[k][i][j]=q;G->img[k][i][j]=V->img[radii_t][i][j];B->img[k][i][j]=p;
	    }else if(Hii==2){
	      R->img[k][i][j]=p;G->img[k][i][j]=V->img[radii_t][i][j];B->img[k][i][j]=t;
	    }else if(Hii==3){
	      R->img[k][i][j]=p;G->img[k][i][j]=q;B->img[k][i][j]=V->img[radii_t][i][j];
	    }else if(Hii==4){
	      R->img[k][i][j]=t;G->img[k][i][j]=p;B->img[k][i][j]=V->img[radii_t][i][j];
	    }else{
	      R->img[k][i][j]=V->img[radii_t][i][j];G->img[k][i][j]=p;B->img[k][i][j]=q;
	    }
	}
	if(R->img[k][i][j]<0.0)R->img[k][i][j]=0.0;
	if(G->img[k][i][j]<0.0)G->img[k][i][j]=0.0;
	if(B->img[k][i][j]<0.0)B->img[k][i][j]=0.0;
	
	if(R->img[k][i][j]>255.0)R->img[k][i][j]=255.0;
	if(G->img[k][i][j]>255.0)G->img[k][i][j]=255.0;
	if(B->img[k][i][j]>255.0)B->img[k][i][j]=255.0;
	
      }
     delete H;
     delete S;
     delete V; 
  }
  

}





/* g++ -o VideoStyle VideoStyle.cxx -lm
引数14個
./VideoStyle 入力フォルダー名 出力フォルダー名 出力ファイル名(.bmpなし) 畳み込みの半径(int) DoGの標準偏差(double) DoGのバンド幅(double) Bilateralフィルタの空間標準偏差(double) Bilateralフィルタの輝度標準偏差(double) Bilateralフィルタの繰り返し回数(int) 時間方向畳み込みの半径(int) Bilateral&DoGフィルタの時間標準偏差(double) HSVのHを多値化する数(int) HSV量子化のV強調パラメータ(double) RGBを多値化する数(int) 
*/

int main(int argc,char *argv[]){
  int sx,sy,st;
  /* 入力3D画像 */
  Image3D *R = NULL;
  Image3D *G = NULL;
  Image3D *B = NULL;
  /* ビデオ(連番画像)の読み込み */
  printf("Start Open Video\n");
  OpenVideo(argv[1],&R,&G,&B,&sx,&sy,&st);
  printf("End Open Video\n");
  int i,j,k;
  /* 計算の為の画像 */
  Image3D *gray = new Image3D(sx,sy,st);
  Image3D *out_gray = new Image3D(sx,sy,st);
  Image3D *outR = new Image3D(sx,sy,st);
  Image3D *outG = new Image3D(sx,sy,st);
  Image3D *outB = new Image3D(sx,sy,st);
  /* 空間畳み込みの半径 */
  int radii = atoi(argv[4]);
  /* DoGの空間標準偏差(double) */
  double sig = atof(argv[5]);
   /* DoGのバンド幅(double) */
  double K = atof(argv[6]);
  /* Bilateralフィルタの空間標準偏差(double) */
  double sigBS = atof(argv[7]);
   /* Bilateralフィルタの輝度標準偏差(double) */
  double sigBI = atof(argv[8]);
  /* Bilateralフィルタの繰り返し回数(int) */
  int N = atoi(argv[9]);
   /* 時間方向畳み込みの半径 */
  int radii_t = atoi(argv[10]);
  /* Bilateral&DoGフィルタの時間標準偏差(double) */
  double sig_t = atof(argv[11]);
   /* HSVのHを多値化する数 */
  int HSV_N = atoi(argv[12]);
  /* HSV量子化のV強調パラメータ */
  double alpha = atof(argv[13]);
  /* RGBを多値化する数 */
  int RGB_N = atoi(argv[14]);
   printf("Start Bilateral Filter\n");
  /* Bilateralフィルタの繰り返し: Bilateral.h */ 
  IterativeBilateralFilter(N,R,G,B,outR,outG,outB,radii,radii_t,sigBS,sig_t,sigBI);
  printf("End Bilateral Filter\n");
  /* outR,outG,outBを0-255へ線形補間で正規化する */
  printf("normalization\n");
  Normalize(outR,outG,outB);
  printf("make gray scale image\n");

   /* グレースケール3D画像の作成 */
  for(k=0;k<st;k++)
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++)
      gray->img[k][i][j] = (outR->img[k][i][j]+outG->img[k][i][j]+outB->img[k][i][j])/3.0;
  
  printf("Quantization\n");
  /* HSV空間での多値化(量子化) */
  QuantizationHSV(outR,outG,outB,HSV_N,radii_t,alpha);
  /* RGB空間での多値化(量子化)　*/
  Quantization(outR,outR,radii_t,RGB_N);
  Quantization(outG,outG,radii_t,RGB_N);
  Quantization(outB,outB,radii_t,RGB_N);
  
  printf("generate DoG\n");
  /* DoG 3D画像の生成 */
  Image3D *dog = makeDoGImage(radii,radii_t,sig,sig_t,K);
  /* DoGの計算 */
  Conv(gray,radii,radii_t,dog,out_gray);
  
  /* Stylizedビデオの合成 */
  for(k=0;k<st;k++)
     for(i=0;i<out_gray->sy;i++)
       for(j=0;j<out_gray->sx;j++){
	 if(out_gray->img[k][i][j]>0.0){
	    /* 繰り返しBilateralフィルタ->HSV多値化->RGB多値化　*/
	   outR->img[k][i][j] = outR->img[k][i][j];
	   outG->img[k][i][j] = outG->img[k][i][j];
	   outB->img[k][i][j] = outB->img[k][i][j];
	   
	 }else{
	   
	    /* 繰り返しBilateralフィルタ->DoGによるEdge */
	   outR->img[k][i][j] = 0.0;
	   outG->img[k][i][j] = 0.0;
	   outB->img[k][i][j] = 0.0;
	   
	 }
	 
      
       }
  
  /* 出力Stylizedビデオの保存 */
  printf("save video\n");
  SaveVideo(argv[2],argv[3],outR,outG,outB);
  
  /* メモリの開放 */
  printf("delete data \n");
  delete dog;
    
  delete gray;
  delete out_gray;
  
  
  delete R;
  delete G;
  delete B;


  delete outR;
  delete outG;
  delete outB;

 
  return 0;
}

#undef PI

